# -*- coding: utf-8 -*-
# Copyright (C) 2014 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
from copy import copy
import re


from tiramisu.i18n import _
from tiramisu.setting import groups, undefined  # , log
from .baseoption import BaseOption, DynSymLinkOption, SymLinkOption, \
    allowed_character
from . import MasterSlaves
from tiramisu.error import ConfigError, ConflictError, ValueWarning
from tiramisu.storage import get_storages_option
from tiramisu.autolib import carry_out_calculation


StorageOptionDescription = get_storages_option('optiondescription')
name_regexp = re.compile(r'^{0}*$'.format(allowed_character))


class OptionDescription(BaseOption, StorageOptionDescription):
    """Config's schema (organisation, group) and container of Options
    The `OptionsDescription` objects lives in the `tiramisu.config.Config`.
    """
    __slots__ = tuple()

    def __init__(self, name, doc, children, requires=None, properties=None):
        """
        :param children: a list of options (including optiondescriptions)

        """
        super(OptionDescription, self).__init__(name, doc=doc,
                                                requires=requires,
                                                properties=properties,
                                                callback=False)
        child_names = []
        dynopt_names = []
        for child in children:
            name = child.impl_getname()
            child_names.append(name)
            if isinstance(child, DynOptionDescription):
                dynopt_names.append(name)

        #better performance like this
        valid_child = copy(child_names)
        valid_child.sort()
        old = None
        for child in valid_child:
            if child == old:  # pragma: optional cover
                raise ConflictError(_('duplicate option name: '
                                      '{0}').format(child))
            if dynopt_names:
                for dynopt in dynopt_names:
                    if child != dynopt and child.startswith(dynopt):
                        raise ConflictError(_('option must not start as '
                                              'dynoptiondescription'))
            old = child
        self._add_children(child_names, children)
        self._cache_consistencies = None
        # the group_type is useful for filtering OptionDescriptions in a config
        self._group_type = groups.default
        self._is_build_cache = False

    def impl_getdoc(self):
        return self.impl_get_information('doc')

    def impl_validate(self, *args):
        """usefull for OptionDescription"""
        pass

    def impl_getpaths(self, include_groups=False, _currpath=None):
        """returns a list of all paths in self, recursively
           _currpath should not be provided (helps with recursion)
        """
        return _impl_getpaths(self, include_groups, _currpath)

    def impl_build_cache_consistency(self, _consistencies=None, cache_option=None):
        if _consistencies is None:
            init = True
            _consistencies = {}
            cache_option = []
        else:
            init = False
        for option in self._impl_getchildren(dyn=False):
            cache_option.append(option._get_id())
            if not isinstance(option, OptionDescription):
                for func, all_cons_opts, params in option._get_consistencies():
                    for opt in all_cons_opts:
                        _consistencies.setdefault(opt,
                                                  []).append((func,
                                                             all_cons_opts,
                                                             params))
            else:
                option.impl_build_cache_consistency(_consistencies, cache_option)
        if init and _consistencies != {}:
            self._cache_consistencies = {}
            for opt, cons in _consistencies.items():
                if opt._get_id() not in cache_option:  # pragma: optional cover
                    raise ConfigError(_('consistency with option {0} '
                                        'which is not in Config').format(
                                            opt.impl_getname()))
                self._cache_consistencies[opt] = tuple(cons)

    def impl_validate_options(self, cache_option=None):
        """validate duplicate option and set option has readonly option
        """
        if cache_option is None:
            init = True
            cache_option = []
        else:
            init = False
        for option in self._impl_getchildren(dyn=False):
            #FIXME specifique id for sqlalchemy?
            #FIXME avec sqlalchemy ca marche le multi parent ? (dans des configs différentes)
            oid = option._get_id()
            cache_option.append(oid)
            option._set_readonly()
            if isinstance(option, OptionDescription):
                option.impl_validate_options(cache_option)
        if init:
            if len(cache_option) != len(set(cache_option)):
                for idx in xrange(1, len(cache_option) + 1):
                    opt = cache_option.pop(0)
                    if opt in cache_option:
                        raise ConflictError(_('duplicate option: {0}').format(opt))
            self._set_readonly()

    # ____________________________________________________________
    def impl_set_group_type(self, group_type):
        """sets a given group object to an OptionDescription

        :param group_type: an instance of `GroupType` or `MasterGroupType`
                              that lives in `setting.groups`
        """
        if self._group_type != groups.default:  # pragma: optional cover
            raise TypeError(_('cannot change group_type if already set '
                            '(old {0}, new {1})').format(self._group_type,
                                                         group_type))
        if isinstance(group_type, groups.GroupType):
            self._group_type = group_type
            if isinstance(group_type, groups.MasterGroupType):
                MasterSlaves(self.impl_getname(), self.impl_getchildren())
        else:  # pragma: optional cover
            raise ValueError(_('group_type: {0}'
                               ' not allowed').format(group_type))

    def _valid_consistency(self, option, value, context, index, submulti_idx):
        if self._cache_consistencies is None:
            return True
        #consistencies is something like [('_cons_not_equal', (opt1, opt2))]
        if isinstance(option, DynSymLinkOption):
            consistencies = self._cache_consistencies.get(option._impl_getopt())
        else:
            consistencies = self._cache_consistencies.get(option)
        if consistencies is not None:
            for func, all_cons_opts, params in consistencies:
                warnings_only = params.get('warnings_only', False)
                transitive = params.get('transitive', True)
                #all_cons_opts[0] is the option where func is set
                if isinstance(option, DynSymLinkOption):
                    subpath = '.'.join(option._dyn.split('.')[:-1])
                    namelen = len(option._impl_getopt().impl_getname())
                    suffix = option.impl_getname()[namelen:]
                    opts = []
                    for opt in all_cons_opts:
                        name = opt.impl_getname() + suffix
                        path = subpath + '.' + name
                        opts.append(opt._impl_to_dyn(name, path))
                else:
                    opts = all_cons_opts
                try:
                    opts[0]._launch_consistency(func, option, value, context,
                                                index, submulti_idx, opts,
                                                warnings_only, transitive)
                except ValueError as err:
                    if warnings_only:
                        raise ValueWarning(err.message, option)
                    else:
                        raise err

    def _impl_getstate(self, descr=None):
        """enables us to export into a dict
        :param descr: parent :class:`tiramisu.option.OptionDescription`
        """
        if descr is None:
            #FIXME faut le desactiver ?
            #self.impl_build_cache_consistency()
            self.impl_build_cache_option()
            descr = self
        super(OptionDescription, self)._impl_getstate(descr)
        self._state_group_type = str(self._group_type)
        for option in self._impl_getchildren():
            option._impl_getstate(descr)

    def __getstate__(self):
        """special method to enable the serialization with pickle
        """
        stated = True
        try:
            # the `_state` attribute is a flag that which tells us if
            # the serialization can be performed
            self._stated
        except AttributeError:
            # if cannot delete, _impl_getstate never launch
            # launch it recursivement
            # _stated prevent __getstate__ launch more than one time
            # _stated is delete, if re-serialize, re-lauch _impl_getstate
            self._impl_getstate()
            stated = False
        return super(OptionDescription, self).__getstate__(stated)

    def _impl_setstate(self, descr=None):
        """enables us to import from a dict
        :param descr: parent :class:`tiramisu.option.OptionDescription`
        """
        if descr is None:
            self._cache_consistencies = None
            self.impl_build_cache_option()
            descr = self
        self._group_type = getattr(groups, self._state_group_type)
        if isinstance(self._group_type, groups.MasterGroupType):
            MasterSlaves(self.impl_getname(), self.impl_getchildren(),
                         validate=False)
        del(self._state_group_type)
        super(OptionDescription, self)._impl_setstate(descr)
        for option in self._impl_getchildren(dyn=False):
            option._impl_setstate(descr)

    def __setstate__(self, state):
        super(OptionDescription, self).__setstate__(state)
        try:
            self._stated
        except AttributeError:
            self._impl_setstate()

    def _impl_get_suffixes(self, context):
        callback, callback_params = self.impl_get_callback()
        values = carry_out_calculation(self, context=context,
                                       callback=callback,
                                       callback_params=callback_params)
        if len(values) > len(set(values)):
            raise ConfigError(_('DynOptionDescription callback return not uniq value'))
        for val in values:
            if not isinstance(val, str) or re.match(name_regexp, val) is None:
                raise ValueError(_("invalid suffix: {0} for option").format(val))
        return values

    def _impl_search_dynchild(self, name=undefined, context=undefined):
        ret = []
        for child in self._impl_st_getchildren(context, only_dyn=True):
            cname = child.impl_getname()
            if name is undefined or name.startswith(cname):
                path = cname
                for value in child._impl_get_suffixes(context):
                    if name is undefined:
                        ret.append(SynDynOptionDescription(child, cname + value, path + value, value))
                    elif name == cname + value:
                        return SynDynOptionDescription(child, name, path + value, value)
        return ret

    def _impl_get_dynchild(self, child, suffix):
        name = child.impl_getname() + suffix
        path = self.impl_getname() + suffix + '.' + name
        if isinstance(child, OptionDescription):
            return SynDynOptionDescription(child, name, path, suffix)
        else:
            return child._impl_to_dyn(name, path)

    def _impl_getchildren(self, dyn=True, context=undefined):
        for child in self._impl_st_getchildren(context):
            cname = child.impl_getname()
            if dyn and child.impl_is_dynoptiondescription():
                path = cname
                for value in child._impl_get_suffixes(context):
                    yield SynDynOptionDescription(child,
                                                  cname + value,
                                                  path + value, value)
            else:
                yield child

    def impl_getchildren(self):
        return list(self._impl_getchildren())

    def __getattr__(self, name, context=undefined):
        if name.startswith('_'):  # or name.startswith('impl_'):
            return object.__getattribute__(self, name)
        if '.' in name:
            path = name.split('.')[0]
            subpath = '.'.join(name.split('.')[1:])
            return self.__getattr__(path, context=context).__getattr__(subpath, context=context)
        return self._getattr(name, context=context)


class DynOptionDescription(OptionDescription):
    def __init__(self, name, doc, children, requires=None, properties=None,
                 callback=None, callback_params=None):
        super(DynOptionDescription, self).__init__(name, doc, children,
                                                   requires, properties)
        for child in children:
            if isinstance(child, OptionDescription):
                if child.impl_get_group_type() != groups.master:
                    raise ConfigError(_('cannot set optiondescription in an '
                                        'dynoptiondescription'))
                for chld in child._impl_getchildren():
                    chld._impl_setsubdyn(self)
            if isinstance(child, SymLinkOption):
                raise ConfigError(_('cannot set symlinkoption in an '
                                    'dynoptiondescription'))
            child._impl_setsubdyn(self)
        self.impl_set_callback(callback, callback_params)
        self.commit()

    def _validate_callback(self, callback, callback_params):
        if callback is None:
            raise ConfigError(_('callback is mandatory for dynoptiondescription'))


class SynDynOptionDescription(object):
    __slots__ = ('_opt', '_name', '_path', '_suffix')

    def __init__(self, opt, name, path, suffix):
        self._opt = opt
        self._name = name
        self._path = path
        self._suffix = suffix

    def __getattr__(self, name, context=undefined):
        if name in dir(self._opt):
            return getattr(self._opt, name)
        return self._opt._getattr(name, suffix=self._suffix, context=context)

    def impl_getname(self):
        return self._name

    def _impl_getchildren(self, dyn=True, context=undefined):
        children = []
        for child in self._opt._impl_getchildren():
            children.append(self._opt._impl_get_dynchild(child, self._suffix))
        return children

    def impl_getchildren(self):
        return self._impl_getchildren()

    def impl_getpath(self, context):
        return self._path

    def impl_getpaths(self, include_groups=False, _currpath=None):
        return _impl_getpaths(self, include_groups, _currpath)

    def _impl_getopt(self):
        return self._opt


def _impl_getpaths(klass, include_groups, _currpath):
        """returns a list of all paths in klass, recursively
           _currpath should not be provided (helps with recursion)
        """
        if _currpath is None:
            _currpath = []
        paths = []
        for option in klass._impl_getchildren():
            attr = option.impl_getname()
            if option.impl_is_optiondescription():
                if include_groups:
                    paths.append('.'.join(_currpath + [attr]))
                paths += option.impl_getpaths(include_groups=include_groups,
                                              _currpath=_currpath + [attr])
            else:
                paths.append('.'.join(_currpath + [attr]))
        return paths
